package java.nio.charset;

import sun.nio.cs.StandardCharsets;
import sun.nio.cs.ThreadLocalCoders;
import java.nio.CharBuffer;
import java.nio.ByteBuffer;
import java.util.Locale;
import java.util.HashSet;
import java.util.Collections;
import java.util.TreeMap;
import sun.misc.ASCIICaseInsensitiveComparator;
import java.util.SortedMap;
import java.util.Map;
import java.security.PrivilegedAction;
import java.util.NoSuchElementException;
import java.util.ServiceConfigurationError;
import java.util.ServiceLoader;
import java.util.Iterator;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import sun.misc.VM;
import java.util.Set;
import java.nio.charset.spi.CharsetProvider;

public abstract class Charset implements Comparable<Charset>
{
    private static volatile String bugLevel;
    private static CharsetProvider standardProvider;
    private static volatile Object[] cache1;
    private static volatile Object[] cache2;
    private static ThreadLocal<ThreadLocal<?>> gate;
    private static volatile Charset defaultCharset;
    private final String name;
    private final String[] aliases;
    private Set<String> aliasSet;
    
    static boolean atBugLevel(String anObject) {
        String bugLevel = Charset.bugLevel;
        if (bugLevel == null) {
            if (!VM.isBooted()) {
                return false;
            }
            bugLevel = (Charset.bugLevel = (String)AccessController.doPrivileged(new GetPropertyAction("sun.nio.cs.bugLevel", "")));
        }
        return bugLevel.equals(anObject);
    }
    
    private static void checkName(String s) {
        int length = s.length();
        if (!atBugLevel("1.4") && length == 0) {
            throw new IllegalCharsetNameException(s);
        }
        for (int i = 0; i < length; ++i) {
            int char1 = s.charAt(i);
            if ((char1 < 'A' || char1 > 'Z') && (char1 < 'a' || char1 > 'z') && (char1 < '0' || char1 > '9') && (char1 != '-' || i == 0) && (char1 != '+' || i == 0) && (char1 != ':' || i == 0) && (char1 != '_' || i == 0) && (char1 != '.' || i == 0)) {
                throw new IllegalCharsetNameException(s);
            }
        }
        return;
    }
    
    private static void cache(String s, Charset charset) {
        Charset.cache2 = Charset.cache1;
        Charset.cache1 = new Object[] { s, charset };
    }
    
    private static Iterator<CharsetProvider> providers() {
        return new Iterator<CharsetProvider>() {
            ClassLoader cl = ClassLoader.getSystemClassLoader();
            ServiceLoader<CharsetProvider> sl = ServiceLoader.load(CharsetProvider.class, this.cl);
            Iterator<CharsetProvider> i = this.sl.iterator();
            CharsetProvider next = null;
            
            private boolean getNext() {
                while (this.next == null) {
                    try {
                        if (!this.i.hasNext()) {
                            return false;
                        }
                        this.next = (CharsetProvider)this.i.next();
                    }
                    catch (ServiceConfigurationError serviceConfigurationError) {
                        if ((serviceConfigurationError.getCause() instanceof SecurityException)) {
                            continue;
                        }
                        throw serviceConfigurationError;
                    }
                }
                return true;
            }
            
            @Override
            public boolean hasNext() {
                return this.getNext();
            }
            
            @Override
            public CharsetProvider next() {
                if (!this.getNext()) {
                    throw new NoSuchElementException();
                }
                CharsetProvider next = this.next;
                this.next = null;
                return next;
            }
            
            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    private static Charset lookupViaProviders(final String s) {
        if (!VM.isBooted()) {
            return null;
        }
        if (Charset.gate.get() != null) {
            return null;
        }
        try {
            Charset.gate.set(Charset.gate);
            return (Charset)AccessController.doPrivileged(new PrivilegedAction<Charset>() {
                @Override
                public Charset run() {
                    Iterator<Object> access$000 = Charset.access$000();
                    while (access$000.hasNext()) {
                        Charset charsetForName = ((CharsetProvider)access$000.next()).charsetForName(s);
                        if (charsetForName != null) {
                            return charsetForName;
                        }
                        continue;
                    }
                    return null;
                }
            });
        }
        finally {
            Charset.gate.set(null);
        }
    }
    
    private static Charset lookupExtendedCharset(String s) {
        CharsetProvider extendedProvider = Charset.ExtendedProviderHolder.extendedProvider;
        return (extendedProvider != null) ? extendedProvider.charsetForName(s) : null;
    }
    
    private static Charset lookup(String s) {
        if (s == null) {
            throw new IllegalArgumentException("Null charset name");
        }
        Object[] cache1;
        if ((cache1 = Charset.cache1) != null && s.equals(cache1[0])) {
            return (Charset)cache1[1];
        }
        return lookup2(s);
    }
    
    private static Charset lookup2(String s) {
        Object[] cache2;
        if ((cache2 = Charset.cache2) != null && s.equals(cache2[0])) {
            Charset.cache2 = Charset.cache1;
            Charset.cache1 = cache2;
            return (Charset)cache2[1];
        }
        Charset charset;
        if ((charset = Charset.standardProvider.charsetForName(s)) != null || (charset = lookupExtendedCharset(s)) != null || (charset = lookupViaProviders(s)) != null) {
            cache(s, charset);
            return charset;
        }
        checkName(s);
        return null;
    }
    
    public static boolean isSupported(String s) {
        return lookup(s) != null;
    }
    
    public static Charset forName(String s) {
        Charset lookup = lookup(s);
        if (lookup != null) {
            return lookup;
        }
        throw new UnsupportedCharsetException(s);
    }
    
    private static void put(Iterator<Charset> iterator, Map<String, Charset> map) {
        while (iterator.hasNext()) {
            Charset charset = (Charset)iterator.next();
            if (!map.containsKey(charset.name())) {
                map.put(charset.name(), charset);
            }
            else {
                continue;
            }
        }
    }
    
    public static SortedMap<String, Charset> availableCharsets() {
        return (SortedMap)AccessController.doPrivileged(new PrivilegedAction<SortedMap<String, Charset>>() {
            @Override
            public SortedMap<String, Charset> run() {
                TreeMap m = new TreeMap(ASCIICaseInsensitiveComparator.CASE_INSENSITIVE_ORDER);
                Charset.access$200(Charset.standardProvider.charsets(), m);
                CharsetProvider extendedProvider = Charset.ExtendedProviderHolder.extendedProvider;
                if (extendedProvider != null) {
                    Charset.access$200(extendedProvider.charsets(), m);
                }
                Iterator<Object> access$000 = Charset.access$000();
                while (access$000.hasNext()) {
                    Charset.access$200(((CharsetProvider)access$000.next()).charsets(), m);
                }
                return Collections.unmodifiableSortedMap(m);
            }
        });
    }
    
    public static Charset defaultCharset() {
        if (Charset.defaultCharset == null) {
            synchronized (Charset.class) {
                Charset lookup = lookup((String)AccessController.doPrivileged(new GetPropertyAction("file.encoding")));
                if (lookup != null) {
                    Charset.defaultCharset = lookup;
                }
                else {
                    Charset.defaultCharset = forName("UTF-8");
                }
            }
        }
        return Charset.defaultCharset;
    }
    
    protected Charset(String name, String[] aliases) {
        this.aliasSet = null;
        checkName(name);
        aliases = ((aliases == null) ? new String[0] : aliases);
        for (int i = 0; i < aliases.length; ++i) {
            checkName(aliases[i]);
        }
        this.name = name;
        this.aliases = aliases;
    }
    
    public final String name() {
        return this.name;
    }
    
    public final Set<String> aliases() {
        if (this.aliasSet != null) {
            return this.aliasSet;
        }
        int length = this.aliases.length;
        HashSet<Object> s = new HashSet(length);
        for (int i = 0; i < length; ++i) {
            s.add(this.aliases[i]);
        }
        return this.aliasSet = Collections.unmodifiableSet(s);
    }
    
    public String displayName() {
        return this.name;
    }
    
    public final boolean isRegistered() {
        return !this.name.startsWith("X-") && !this.name.startsWith("x-");
    }
    
    public String displayName(Locale locale) {
        return this.name;
    }
    
    public abstract boolean contains(Charset p0);
    
    public abstract CharsetDecoder newDecoder();
    
    public abstract CharsetEncoder newEncoder();
    
    public boolean canEncode() {
        return true;
    }
    
    public final CharBuffer decode(ByteBuffer byteBuffer) {
        try {
            return ThreadLocalCoders.decoderFor(this).onMalformedInput(CodingErrorAction.REPLACE).onUnmappableCharacter(CodingErrorAction.REPLACE).decode(byteBuffer);
        }
        catch (CharacterCodingException cause) {
            throw new Error(cause);
        }
    }
    
    public final ByteBuffer encode(CharBuffer charBuffer) {
        try {
            return ThreadLocalCoders.encoderFor(this).onMalformedInput(CodingErrorAction.REPLACE).onUnmappableCharacter(CodingErrorAction.REPLACE).encode(charBuffer);
        }
        catch (CharacterCodingException cause) {
            throw new Error(cause);
        }
    }
    
    public final ByteBuffer encode(String csq) {
        return this.encode(CharBuffer.wrap(csq));
    }
    
    @Override
    public final int compareTo(Charset charset) {
        return this.name().compareToIgnoreCase(charset.name());
    }
    
    @Override
    public final int hashCode() {
        return this.name().hashCode();
    }
    
    @Override
    public final boolean equals(Object o) {
        return o instanceof Charset && (this == o || this.name.equals(((Charset)o).name()));
    }
    
    @Override
    public final String toString() {
        return this.name();
    }
    
    static Iterator access$000() {
        return providers();
    }
    
    static void access$200(Iterator iterator, Map map) {
        put(iterator, map);
    }
    
    static {
        Charset.bugLevel = null;
        Charset.standardProvider = new StandardCharsets();
        Charset.cache1 = null;
        Charset.cache2 = null;
        Charset.gate = new ThreadLocal();
    }
    
    private static class ExtendedProviderHolder
    {
        static final CharsetProvider extendedProvider;
        
        private static CharsetProvider extendedProvider() {
            return (CharsetProvider)AccessController.doPrivileged(new PrivilegedAction<CharsetProvider>() {
                @Override
                public CharsetProvider run() {
                    try {
                        return (CharsetProvider)Class.forName("sun.nio.cs.ext.ExtendedCharsets").newInstance();
                    }
                    catch (ClassNotFoundException ex) {}
                    catch (InstantiationException | IllegalAccessException cause) {
                        throw new Error(cause);
                    }
                    return null;
                }
            });
        }
        
        static {
            extendedProvider = extendedProvider();
        }
    }
}
